import json
import os

from pydantic import BaseModel

from marker.renderers.html import HTMLOutput
from marker.renderers.json import JSONOutput
from marker.renderers.markdown import MarkdownOutput
from marker.settings import settings


def output_exists(output_dir: str, fname_base: str):
    exts = ["md", "html", "json"]
    for ext in exts:
        if os.path.exists(os.path.join(output_dir, f"{fname_base}.{ext}")):
            return True
    return False


def text_from_rendered(rendered: BaseModel):
    if isinstance(rendered, MarkdownOutput):
        return rendered.markdown, "md", rendered.images
    elif isinstance(rendered, HTMLOutput):
        return rendered.html, "html", rendered.images
    elif isinstance(rendered, JSONOutput):
        return rendered.model_dump_json(exclude=["metadata"], indent=2), "json", {}
    else:
        raise ValueError("Invalid output type")


def save_output(rendered: BaseModel, output_dir: str, fname_base: str):
    text, ext, images = text_from_rendered(rendered)
    text = text.encode(settings.OUTPUT_ENCODING, errors='replace').decode(settings.OUTPUT_ENCODING)

    with open(os.path.join(output_dir, f"{fname_base}.{ext}"), "w+", encoding=settings.OUTPUT_ENCODING) as f:
        f.write(text)
    with open(os.path.join(output_dir, f"{fname_base}_meta.json"), "w+", encoding=settings.OUTPUT_ENCODING) as f:
        f.write(json.dumps(rendered.metadata, indent=2))

    for img_name, img in images.items():
        img.save(os.path.join(output_dir, img_name), settings.OUTPUT_IMAGE_FORMAT)



from typing import Annotated, List, Optional, Tuple

from texify.inference import batch_inference
from texify.model.model import GenerateVisionEncoderDecoderModel
from tqdm import tqdm

from marker.models import TexifyPredictor
from marker.processors import BaseProcessor
from marker.schema import BlockTypes
from marker.schema.document import Document
from marker.settings import settings


class EquationProcessor(BaseProcessor):
    """
    A processor for recognizing equations in the document.
    """
    block_types: Annotated[
        Tuple[BlockTypes],
        "The block types to process.",
    ] = (BlockTypes.Equation,)
    model_max_length: Annotated[
        int,
        "The maximum number of tokens to allow for the Texify model.",
    ] = 384
    texify_batch_size: Annotated[
        Optional[int],
        "The batch size to use for the Texify model.",
        "Default is None, which will use the default batch size for the model."
    ] = None
    token_buffer: Annotated[
        int,
        "The number of tokens to buffer above max for the Texify model.",
    ] = 256

    def __init__(self, texify_model: TexifyPredictor, config=None):
        super().__init__(config)

        self.texify_model = texify_model

    def __call__(self, document: Document):
        equation_data = []

        for page in document.pages:
            for block in page.contained_blocks(document, self.block_types):
                image = block.get_image(document, highres=False).convert("RGB")
                raw_text = block.raw_text(document)
                token_count = self.get_total_texify_tokens(raw_text)

                equation_data.append({
                    "image": image,
                    "block_id": block.id,
                    "token_count": token_count
                })

        predictions = self.get_latex_batched(equation_data)
        for prediction, equation_d in zip(predictions, equation_data):
            conditions = [
                self.get_total_texify_tokens(prediction) < self.model_max_length,
                # Make sure we didn't get to the overall token max, indicates run-on
                len(prediction) > equation_d["token_count"] * .4,
                len(prediction.strip()) > 0
            ]
            if not all(conditions):
                continue

            block = document.get_block(equation_d["block_id"])
            block.html = self.parse_latex_to_html(prediction)

    def parse_latex_to_html(self, latex: str):
        html_out = ""
        try:
            latex = self.parse_latex(latex)
        except ValueError as e:
            # If we have mismatched delimiters, we'll treat it as a single block
            # Strip the $'s from the latex
            latex = [
                {"class": "block", "content": latex.replace("$", "")}
            ]

        for el in latex:
            if el["class"] == "block":
                html_out += f'<math display="block">{el["content"]}</math>'
            elif el["class"] == "inline":
                html_out += f'<math display="inline">{el["content"]}</math>'
            else:
                html_out += f" {el['content']} "
        return html_out.strip()

    def get_batch_size(self):
        if self.texify_batch_size is not None:
            return self.texify_batch_size
        elif settings.TORCH_DEVICE_MODEL == "cuda":
            return 6
        elif settings.TORCH_DEVICE_MODEL == "mps":
            return 6
        return 2

    def get_latex_batched(self, equation_data: List[dict]):
        predictions = [""] * len(equation_data)
        batch_size = self.get_batch_size()

        for i in tqdm(range(0, len(equation_data), batch_size), desc="Recognizing equations"):
            # Dynamically set max length to save inference time
            min_idx = i
            max_idx = min(min_idx + batch_size, len(equation_data))

            batch_equations = equation_data[min_idx:max_idx]
            max_length = max([eq["token_count"] for eq in batch_equations])
            max_length = min(max_length, self.model_max_length)
            max_length += self.token_buffer

            batch_images = [eq["image"] for eq in batch_equations]

            model_output = self.texify_model(
                batch_images,
                max_tokens=max_length
            )

            for j, output in enumerate(model_output):
                token_count = self.get_total_texify_tokens(output)
                if token_count >= max_length - 1:
                    output = ""

                image_idx = i + j
                predictions[image_idx] = output
        return predictions

    def get_total_texify_tokens(self, text):
        tokenizer = self.texify_model.processor.tokenizer
        tokens = tokenizer(text)
        return len(tokens["input_ids"])


    @staticmethod
    def parse_latex(text: str):
        if text.count("$") % 2 != 0:
            raise ValueError("Mismatched delimiters in LaTeX")

        DELIMITERS = [
            ("$$", "block"),
            ("$", "inline")
        ]

        text = text.replace("\n", "<br>")  # we can't handle \n's inside <p> properly if we don't do this

        i = 0
        stack = []
        result = []
        buffer = ""

        while i < len(text):
            for delim, class_name in DELIMITERS:
                if text[i:].startswith(delim):
                    if stack and stack[-1] == delim:  # Closing
                        stack.pop()
                        result.append({"class": class_name, "content": buffer})
                        buffer = ""
                        i += len(delim)
                        break
                    elif not stack:  # Opening
                        if buffer:
                            result.append({"class": "text", "content": buffer})
                        stack.append(delim)
                        buffer = ""
                        i += len(delim)
                        break
                    else:
                        raise ValueError(f"Nested {class_name} delimiters not supported")
            else:  # No delimiter match
                buffer += text[i]
                i += 1

        if buffer:
            result.append({"class": "text", "content": buffer})
        return result



from statistics import mean

from marker.processors import BaseProcessor
from marker.schema import BlockTypes
from marker.schema.document import Document


class OrderProcessor(BaseProcessor):
    """
    A processor for sorting the blocks in order if needed.  This can help when the layout image was sliced.
    """
    block_types = tuple()

    def __call__(self, document: Document):
        for page in document.pages:
            if page.text_extraction_method != "pdftext":
                continue

            if not page.layout_sliced:
                continue

            block_idxs = {}
            for block_id in page.structure:
                block = document.get_block(block_id)
                spans = block.contained_blocks(document, (BlockTypes.Span, ))
                if len(spans) == 0:
                    continue

                block_idxs[block_id] = (spans[0].minimum_position + spans[-1].maximum_position) / 2

            for block_id in page.structure:
                if block_id in block_idxs and block_idxs[block_id] > 0:
                    continue
                block = document.get_block(block_id)
                prev_block = document.get_prev_block(block)
                next_block = document.get_next_block(block)

                while prev_block and prev_block.id not in block_idxs:
                    prev_block = document.get_prev_block(prev_block)

                if not prev_block:
                    while next_block and next_block.id not in block_idxs:
                        next_block = document.get_next_block(next_block)

                if not next_block and not prev_block:
                    block_idxs[block_id] = 0
                elif prev_block:
                    block_idxs[block_id] = block_idxs[prev_block.id] + 1
                else:
                    block_idxs[block_id] = block_idxs[next_block.id] - 1

            page.structure = sorted(page.structure, key=lambda x: block_idxs[x])




import os

from marker.settings import settings

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1" # Transformers uses .isin for a simple op, which is not supported on MPS

from typing import List
from PIL import Image

from surya.detection import DetectionPredictor
from surya.layout import LayoutPredictor
from surya.ocr_error import OCRErrorPredictor
from surya.recognition import RecognitionPredictor
from surya.table_rec import TableRecPredictor

from texify.model.model import load_model as load_texify_model
from texify.model.processor import load_processor as load_texify_processor
from texify.inference import batch_inference

class TexifyPredictor:
    def __init__(self, device=None, dtype=None):
        if not device:
            device = settings.TORCH_DEVICE_MODEL
        if not dtype:
            dtype = settings.TEXIFY_DTYPE

        self.model = load_texify_model(checkpoint=settings.TEXIFY_MODEL_NAME, device=device, dtype=dtype)
        self.processor = load_texify_processor()
        self.device = device
        self.dtype = dtype

    def __call__(self, batch_images: List[Image.Image], max_tokens: int):
        return batch_inference(
            batch_images,
            self.model,
            self.processor,
            max_tokens=max_tokens
        )


def create_model_dict(device=None, dtype=None) -> dict:
    return {
        "layout_model": LayoutPredictor(device=device, dtype=dtype),
        "texify_model": TexifyPredictor(device=device, dtype=dtype),
        "recognition_model": RecognitionPredictor(device=device, dtype=dtype),
        "table_rec_model": TableRecPredictor(device=device, dtype=dtype),
        "detection_model": DetectionPredictor(device=device, dtype=dtype),
        "ocr_error_model": OCRErrorPredictor(device=device, dtype=dtype)
    }


import json
import time
from typing import List

import PIL
import google.generativeai as genai
from google.ai.generativelanguage_v1beta.types import content
from google.api_core.exceptions import ResourceExhausted

from marker.schema.blocks import Block


class GoogleModel:
    def __init__(self, api_key: str, model_name: str):
        if api_key is None:
            raise ValueError("Google API key is not set")

        self.api_key = api_key
        self.model_name = model_name
        self.model = self.configure_google_model()

    def configure_google_model(self):
        genai.configure(api_key=self.api_key)
        return genai.GenerativeModel(self.model_name)

    def generate_response(
            self,
            prompt: str,
            image: PIL.Image.Image | List[PIL.Image.Image],
            block: Block,
            response_schema: content.Schema,
            max_retries: int = 3,
            timeout: int = 60
    ):
        if not isinstance(image, list):
            image = [image]
        tries = 0
        while tries < max_retries:
            try:
                responses = self.model.generate_content(
                    image + [prompt], # According to gemini docs, it performs better if the image is the first element
                    stream=False,
                    generation_config={
                        "temperature": 0,
                        "response_schema": response_schema,
                        "response_mime_type": "application/json",
                    },
                    request_options={'timeout': timeout}
                )
                output = responses.candidates[0].content.parts[0].text
                total_tokens = responses.usage_metadata.total_token_count
                block.update_metadata(llm_tokens_used=total_tokens, llm_request_count=1)
                return json.loads(output)
            except ResourceExhausted as e:
                tries += 1
                wait_time = tries * 3
                print(f"ResourceExhausted: {e}. Retrying in {wait_time} seconds... (Attempt {tries}/{max_retries})")
                time.sleep(wait_time)
            except Exception as e:
                print(e)
                break

        return {}